#!/usr/bin/env python

# DEMO PROGRAM FOR EX9000M/D SERIES MODULES
# DESIGNED ON EX9055M
# LOGIC SUPPLY, 15 Jul 2015

from serial import Serial
import time
import sys

ITER_DELAY = 0.4
"""Convert to ascii hex the DAQ module can understand
Strips leading '0x' and left zero-fills to [fill] digits"""
def hex_val_only(num, fill=2):
    return hex(num)[2:].upper().zfill(fill)

"""Make a serial packet
GENERAL PACKET STRUCTURE
    Delimiter, Address, Command, [Data], [Checksum], Carriage Return"""
def serial_packet(delim, addr, command, *args):
    return ''.join((delim, hex_val_only(addr), command) + args + ('\r\n',))

"""Usage notes"""
def print_usage():
    print "Usage: topsccc_demo.py PORT [--demo|--term-normal|--term-modbus|" + \
          "\t                                               --modbus|--normal]\r\n" + \
          "\t--demo          Run the demo program\r\n" + \
          "\t--term-normal   Normal prompt- enter and execute commands as documented\r\n" + \
          "\t--term-modbus   Modbus prompt- enter and execute commands as documented\r\n" + \
          "\t--modbus        Switch module operating mode to modbus mode\r\n" + \
          "\t--normal        Switch module operating mode to normal mode\r\n"

"""wrapper for port.write so we can see what's happening"""
def write(data):
    if type(data) is list:
        print "Modbus packet: ",
    else:
        print "Normal packet: ",
    print data[:-1]
    port.write(data)

"""nice packet-making wrapper for modbus"""
def packetize_CRC(data):
    return data + modbusCRC(data)

"""modbus crc generator - crc is required for module modbus protocol"""
def modbusCRC(data):
    # CRC Generation
    # 1. Take 0xFFFF, XOR with the first byte in packet
    # 2. If LSB of result is 1, you will need to XOR a second value after performing step 3
    # 3. Shift result 1 place right.
    # 4. If the LSB at step 2 was 1, XOR the result with 0xA001
    # 5. Repeat steps 2-4 7 more times
    # 6. Repeat steps 2-5 for all bytes in packet
    # 7. The resulting CRC is 2 bytes. When sending the packet, the least significant BYTE
    #    is sent first, followed by the most significant BYTE.
    # result with the carry 0xA001
    crc = 0xFFFF # start off with 0xFFFF
    for byte in data: 
        crc = crc ^ (0x0000 | byte) # XOR each byte against existing CRC
        for i in range(8):
            carry = 0xA001 if crc & 0x0001 else 0x0000 # check LSB
            crc = (crc >> 1) ^ carry        #bit-shift right and XOR with 0xA001 if LSB eq'd 1
        crcl = crc % 256 # isolate LSB
        crch = (crc - crcl) >> 8 # shift MSB down 1 byte
    return [crcl, crch] # swap bytes and return
    
def normal_demo_9055(address):
    #get input status - ex. $006\r\n
    write(serial_packet('$', address, '6'))
    print "\nInput status: %s" % port.read(100)[:-1]
    # MANIPULATING OUTPUTS:
    # COMMAND: #[Address][Channel][State]
    # To manipulate individual outputs, channel should be 1c, where c is the
    # desired output, starting from 0 (ex. 11 for the second output). State
    # in this case is either 1 or 0, all other values are ignored
    # To change all outputs at once, channel should be set to 00. Each bit
    # in state corresponds to an output, where the least significant bit is
    # output 0 (ex. 001A will turn out outputs 1, 3, and 4)
    print "\nClearing outputs:"
    #clear all outputs: #000000\r\n
    write(serial_packet('#', address, hex_val_only(0), hex_val_only(0)))
    time.sleep(0.1)
    #loop through output states: #001001\r\n, #001101\r\n, ...
    print "\nSetting outputs one at a time:"
    for i in range(0, 8):
        pack = serial_packet('#', address, '1%s' % i, hex_val_only(1))
        write(pack)
        time.sleep(ITER_DELAY)
        if not port.read(2)[0] == '>':
            print 'Command failed'
    #clear all outputs
    print "\nClearing outputs"
    write(serial_packet('#', address, hex_val_only(0), hex_val_only(0)))
    print port.read(2)
    time.sleep(0.05)
    #turn on, outputs 0, 1, and 3: #00001A\r\n
    #then walk those bits through channels
    print "\nSetting outputs 1, 3, 4"
    for i in range(9):
        write(serial_packet('#', address, hex_val_only(0), hex_val_only(0x00FF & (0x0B << i))))
        time.sleep(ITER_DELAY)
    port.read(2)

def normal_demo_9060(address):
    write(serial_packet('$', address, '6'))
    print "\nInput status: %s" % port.read(100)[:-1]
    # MANIPULATING OUTPUTS:
    # COMMAND: #[Address][Channel][State]
    # To manipulate individual outputs, channel should be 1c, where c is the
    # desired output, starting from 0 (ex. 11 for the second output). State
    # in this case is either 1 or 0, all other values are ignored
    # To change all outputs at once, channel should be set to 00. Each bit
    # in state corresponds to an output, where the least significant bit is
    # output 0 (ex. 001A will turn out outputs 1, 3, and 4)
    print "\nClearing outputs:"
    #clear all outputs: #000000\r\n
    write(serial_packet('#', address, hex_val_only(0), hex_val_only(0)))
    time.sleep(0.1)
    #loop through output states: #001001\r\n, #001101\r\n, ...
    print "\nSetting outputs one at a time:"
    for i in range(0, 4):
        pack = serial_packet('#', address, '1%s' % i, hex_val_only(1))
        write(pack)
        time.sleep(ITER_DELAY)
        if not port.read(2)[0] == '>':
            print 'Command failed'
    #clear all outputs
    print "\nClearing outputs"
    write(serial_packet('#', address, hex_val_only(0), hex_val_only(0)))
    print port.read(2)
    time.sleep(0.05)
    #turn on, outputs 0, 1, and 3: #00001A\r\n
    #then walk those bits through channels
    print "\nSetting outputs 1, 3, 4"
    for i in range(5):
        write(serial_packet('#', address, hex_val_only(0), hex_val_only(0x00FF & (0x0B << i))))
        time.sleep(ITER_DELAY)
    port.read(2)

def normal_demo_9080(address):
    print ''
    for i in [0, 1]:
        print "Reading counter %s: " % i,
        write(serial_packet('#', address, hex_val_only(i, 1)))
        result = port.read(10)
        if result and result[0] == '>':
            print result[3:]
    print "\nClearing outputs:"
    write(serial_packet('@', address, 'DO', hex_val_only(0)))
    print "Walking outputs:"
    for i in range(0, 4):
        write(serial_packet('@', address, 'DO', hex_val_only(i)))
        time.sleep(ITER_DELAY)
        if not port.read(4)[0] == '!':
            print 'Command failed'

def normal_demo_9017(address):
    print "Enabling all channels."
    write(serial_packet('$', address, '5', 'FF'))
    print "Reading analog inputs: ",
    write(serial_packet('#', address))
    reply = port.read(100) # no documented size of return packet
    if reply and reply[0] == '>':
        print reply
    else:
        print ''

def modbus_demo(address):
    print "Running modbus demo..."
    port.flushInput()
    # Read input format:
    # Address,
    # Input command (0x02),
    # Starting channel upper byte,
    # Starting channel lower byte,
    # Quantity of channels to read upper byte,
    # Quantity of channels to read lower byte
    write(packetize_CRC([address, 0x02, 0x00, 0x00, 0x00, 0x08]))
    #write(packetize_CRC([address, 0x02, 0x00, 0x00, 0x00, 0x04])) # for EX9060
    print "\nInput status: ",
    for char in port.read(5):
        print hex(ord(char))[2:].zfill(2).upper(),
    print "\n\nClearing outputs"
    # Write multi output format:
    # Address,
    # Multi-output command (0x0F),
    # Starting channel upper byte,
    # Starting channel lower byte,
    # Quantity of channels to read upper byte,
    # Quantity of channels to read lower byte,
    # Number of output state bytes
    # Output state byte(s) (each bit corresponds to a channel)
    write(packetize_CRC([address, 0x0F, 0x00, 0x00, 0x00, 0x08, 0x01, 0x00]))
    #write(packetize_CRC([address, 0x0F, 0x00, 0x00, 0x00, 0x08, 0x01, 0x00]))
    time.sleep(0.05)
    print "\nSetting outputs one at a time:"
    for i in range(8):
        # Write single output format:
        # Address,
        # Single-output command (0x05),
        # Channel number upper byte,
        # Channel number lower byte,
        # Channel state upper byte (0x00 or 0xFF),
        # Channel state lower byte (always 0x00)
        write(packetize_CRC([address, 0x05, 0x00, i, 0xFF, 0x00]))
        time.sleep(ITER_DELAY)
    print "\nWalk 1011 through channels"
    for i in range(9):
        write(packetize_CRC([address, 0x0F, 0x00, 0x00, 0x00, 0x08, 0x01, 0x00FF & (0x0B << i)]))
        time.sleep(ITER_DELAY)
    port.flushInput() # since we haven't read responses, there's some buildup in the rs485 input buffer

# First, we initialize our RS485 port. For convenience's sake, we require you
# to provide it when you call the demo - demo.py PORT
if not len(sys.argv) > 1 or "--help" in sys.argv:
    print_usage()
    sys.exit(0)

port = Serial(sys.argv[1], baudrate=9600, timeout=0.1)

# Check for demo launch parameters
if len(sys.argv) <= 2 or sys.argv[2] == '--demo':
    print "Searching for modules..."
    address = None
    for i in range(0, 256):
        #check for normal module
        write(serial_packet('$', i, 'M'))
        result = port.read(100)
        if result and result[0] == '!':
            module = result[3:-1]
            address = i
            print "Found module %s at address %s" % (module, address)
            break
        write(packetize_CRC([i, 0x02, 0x00, 0x00, 0x00, 0x01]))
        result = port.read(5)
        if result:
            module = 'modbus'
            address = i
            print "Found modbus module at address %s" % address
            break
        
    # The simplest way to interact with the modules via software is hard-coding
    # the packets (though obviously not the best). It allows us to demonstrate
    # some simple tests without a great deal of overhead.
    if not address is None:
        if module == '9055M' or module == '9044D':
            normal_demo_9055(address)
        elif module == '9050M':
            normal_demo_9050(address)
        elif module == '9060M':
            normal_demo_9060(address)
        elif module == '9080M':
            normal_demo_9080(address)
        elif module == '9017M':
            normal_demo_9017(address)
        elif module == 'modbus':
            modbus_demo(address)
        sys.exit(0)
    else:
        print "No module found."
        sys.exit(0)
elif sys.argv[2] == '--term-normal':
    #terminal
    print "\nEntering demo terminal mode.\nEnter commands as documented and press Enter to send.\nPress CTRL+C to exit.\n"
    try:
        while 1:
            command = ''.join((raw_input('Command: '), '\r\n',))
            port.write(command)
            print '  Reply: %s' % port.read(100)[:-1]
    except KeyboardInterrupt:
        print "\nExiting demo terminal."
    sys.exit(0)
elif sys.argv[2] == '--term-modbus':
    print "\nEntering demo modbus terminal mode.\nEnter commands as space-separated hex bytes (0x00 0x01 ...) and \npress Enter to send.\nPress CRTL+C to exit.\n"
    while 1:
        command = []
        for byte in raw_input("Command: ").split(' '):
            command.append(int(byte[2:], 16))
        port.write(packetize_CRC(command))
        print "Reply:  ",
        for char in port.read(50):
            print "0x%s" % hex(ord(char))[2:].zfill(2),
        print ""
elif sys.argv[2] == '--modbus':
    # switch to modbus
    if len(sys.argv) >= 4:
        write(serial_packet('$', int(sys.argv[3]), 'P', '1'))
        result = port.read(100)
        if result[0] == '!':
            print 'Update successful! Restart to operate in modbus mode.'
        elif result[0] == '?':
            print 'Update failed! Module returned invalid command.'
        else:
            print 'Update failed! No response from module at address %s.' % sys.argv[3]
    else:
        print 'You must provide a module address (topsccc.py PORT --modbus ADDRESS).'
    sys.exit(0)
elif sys.argv[2] == '--normal':
    # switch to normal
    if len(sys.argv) >= 4:
        write(serial_packet('$', int(sys.argv[3]), 'P', '0'))
        result = port.read(100)
        if result and result[0] == '!':
            print 'Update successful! Restart to operate in normal mode.'
        elif result and result[0] == '?':
            print 'Update failed! Module returned invalid command.'
        else:
            print 'Update failed! No response from module at address %s.' % sys.argv[3]
    else:
        print 'You must provide a module address (topsccc.py PORT --normal ADDRESS).'
    sys.exit(0)
else:
    print_usage()



